@@ -1,5 +1,5 @@
 /*
- * This file is part of the Micro Python project, http://micropython.org/
+ * This file is part of the MicroPython project, http://micropython.org/
  *
  * The MIT License (MIT)
  *
@@ -36,60 +36,6 @@ typedef union {
     };
 } float_s_t;
 
-typedef union {
-    double d;
-    struct {
-        uint64_t m : 52;
-        uint64_t e : 11;
-        uint64_t s : 1;
-    };
-} double_s_t;
-
-double __attribute__((pcs("aapcs"))) __aeabi_i2d(int32_t x) {
-    return (float)x;
-}
-
-// TODO
-long long __attribute__((pcs("aapcs"))) __aeabi_f2lz(float x) {
-    return (long)x;
-}
-
-double __attribute__((pcs("aapcs"))) __aeabi_f2d(float x) {
-    float_s_t fx={0};
-    double_s_t dx={0};
-
-    fx.f = x;
-    dx.s = (fx.s);
-    dx.e = (fx.e-127+1023) & 0x7FF;
-    dx.m = fx.m;
-    dx.m <<=(52-23); // left justify
-    return dx.d;
-}
-
-float __attribute__((pcs("aapcs"))) __aeabi_d2f(double x) {
-    float_s_t fx={0};
-    double_s_t dx={0};
-
-    dx.d = x;
-    fx.s = (dx.s);
-    fx.e = (dx.e-1023+127) & 0xFF;
-    fx.m = (dx.m>>(52-23)); // right justify
-    return fx.f;
-}
-
-double __aeabi_dmul(double x , double y) {
-    return 0.0;
-
-}
-
-float sqrtf(float x) {
-    asm volatile (
-            "vsqrt.f32  %[r], %[x]\n"
-            : [r] "=t" (x)
-            : [x] "t"  (x));
-    return x;
-}
-
 #ifndef NDEBUG
 float copysignf(float x, float y) {
     float_s_t fx={.f = x};
@@ -102,27 +48,15 @@ float copysignf(float x, float y) {
 }
 #endif
 
-// some compilers define log2f in terms of logf
-#ifdef log2f
-#undef log2f
-#endif
-// some compilers have _M_LN2 defined in math.h, some don't
-#ifndef _M_LN2
-#define _M_LN2 (0.69314718055994530942)
-#endif
-float log2f(float x) { return logf(x) / (float)_M_LN2; }
-
 static const float _M_LN10 = 2.30258509299404; // 0x40135d8e
 float log10f(float x) { return logf(x) / (float)_M_LN10; }
 
-float tanhf(float x) { return sinhf(x) / coshf(x); }
-
-// TODO we need import these functions from some library (eg musl or newlib)
-float tgammaf(float x) { return 0.0; }
-float lgammaf(float x) { return 0.0; }
-float erff(float x) { return 0.0; }
-float erfcf(float x) { return 0.0; }
-float ldexpf(float x, int exp) { return 0.0; }
+float tanhf(float x) {
+    if (isinf(x)) {
+        return copysignf(1, x);
+    }
+    return sinhf(x) / coshf(x);
+}
 
 /*****************************************************************************/
 /*****************************************************************************/
