@@ -1,5 +1,5 @@
 /*
- * This file is part of the Micro Python project, http://micropython.org/
+ * This file is part of the MicroPython project, http://micropython.org/
  *
  * The MIT License (MIT)
  *
@@ -27,9 +27,6 @@
 
 #include <string.h>
 
-#include "py/nlr.h"
-#include "py/obj.h"
-#include "py/runtime0.h"
 #include "py/runtime.h"
 
 // Helpers for sequence types
@@ -38,9 +35,9 @@
 
 // Implements backend of sequence * integer operation. Assumes elements are
 // memory-adjacent in sequence.
-void mp_seq_multiply(const void *items, mp_uint_t item_sz, mp_uint_t len, mp_uint_t times, void *dest) {
-    for (mp_uint_t i = 0; i < times; i++) {
-        uint copy_sz = item_sz * len;
+void mp_seq_multiply(const void *items, size_t item_sz, size_t len, size_t times, void *dest) {
+    for (size_t i = 0; i < times; i++) {
+        size_t copy_sz = item_sz * len;
         memcpy(dest, items, copy_sz);
         dest = (char*)dest + copy_sz;
     }
@@ -53,51 +50,80 @@ bool mp_seq_get_fast_slice_indexes(mp_uint_t len, mp_obj_t slice, mp_bound_slice
     mp_int_t start, stop;
     mp_obj_slice_get(slice, &ostart, &ostop, &ostep);
 
-    if (ostart == mp_const_none) {
-        start = 0;
+    if (ostep != mp_const_none && ostep != MP_OBJ_NEW_SMALL_INT(1)) {
+        indexes->step = mp_obj_get_int(ostep);
+        if (indexes->step == 0) {
+            mp_raise_ValueError("slice step cannot be zero");
+        }
     } else {
-        start = MP_OBJ_SMALL_INT_VALUE(ostart);
+        indexes->step = 1;
+    }
+
+    if (ostart == mp_const_none) {
+        if (indexes->step > 0) {
+            start = 0;
+        } else {
+            start = len - 1;
+        }
+    } else {
+        start = mp_obj_get_int(ostart);
     }
     if (ostop == mp_const_none) {
-        stop = len;
+        if (indexes->step > 0) {
+            stop = len;
+        } else {
+            stop = 0;
+        }
     } else {
-        stop = MP_OBJ_SMALL_INT_VALUE(ostop);
+        stop = mp_obj_get_int(ostop);
+        if (stop >= 0 && indexes->step < 0) {
+            stop += 1;
+        }
     }
 
     // Unlike subscription, out-of-bounds slice indexes are never error
     if (start < 0) {
         start = len + start;
         if (start < 0) {
-            start = 0;
+            if (indexes->step < 0) {
+                start = -1;
+            } else {
+                start = 0;
+            }
         }
-    } else if ((mp_uint_t)start > len) {
+    } else if (indexes->step > 0 && (mp_uint_t)start > len) {
         start = len;
+    } else if (indexes->step < 0 && (mp_uint_t)start >= len) {
+        start = len - 1;
     }
     if (stop < 0) {
         stop = len + stop;
+        if (stop < 0) {
+            stop = -1;
+        }
+        if (indexes->step < 0) {
+            stop += 1;
+        }
     } else if ((mp_uint_t)stop > len) {
         stop = len;
     }
 
     // CPython returns empty sequence in such case, or point for assignment is at start
-    if (start > stop) {
+    if (indexes->step > 0 && start > stop) {
         stop = start;
+    } else if (indexes->step < 0 && start < stop) {
+        stop = start + 1;
     }
 
     indexes->start = start;
     indexes->stop = stop;
 
-    if (ostep != mp_const_none && ostep != MP_OBJ_NEW_SMALL_INT(1)) {
-        indexes->step = MP_OBJ_SMALL_INT_VALUE(ostep);
-        return false;
-    }
-    indexes->step = 1;
-    return true;
+    return indexes->step == 1;
 }
 
 #endif
 
-mp_obj_t mp_seq_extract_slice(mp_uint_t len, const mp_obj_t *seq, mp_bound_slice_t *indexes) {
+mp_obj_t mp_seq_extract_slice(size_t len, const mp_obj_t *seq, mp_bound_slice_t *indexes) {
     (void)len; // TODO can we remove len from the arg list?
 
     mp_int_t start = indexes->start, stop = indexes->stop;
@@ -106,10 +132,9 @@ mp_obj_t mp_seq_extract_slice(mp_uint_t len, const mp_obj_t *seq, mp_bound_slice
     mp_obj_t res = mp_obj_new_list(0, NULL);
 
     if (step < 0) {
-        stop--;
-        while (start <= stop) {
-            mp_obj_list_append(res, seq[stop]);
-            stop += step;
+        while (start >= stop) {
+            mp_obj_list_append(res, seq[start]);
+            start += step;
         }
     } else {
         while (start < stop) {
@@ -122,7 +147,7 @@ mp_obj_t mp_seq_extract_slice(mp_uint_t len, const mp_obj_t *seq, mp_bound_slice
 
 // Special-case comparison function for sequences of bytes
 // Don't pass MP_BINARY_OP_NOT_EQUAL here
-bool mp_seq_cmp_bytes(mp_uint_t op, const byte *data1, mp_uint_t len1, const byte *data2, mp_uint_t len2) {
+bool mp_seq_cmp_bytes(mp_uint_t op, const byte *data1, size_t len1, const byte *data2, size_t len2) {
     if (op == MP_BINARY_OP_EQUAL && len1 != len2) {
         return false;
     }
@@ -130,14 +155,14 @@ bool mp_seq_cmp_bytes(mp_uint_t op, const byte *data1, mp_uint_t len1, const byt
     // Let's deal only with > & >=
     if (op == MP_BINARY_OP_LESS || op == MP_BINARY_OP_LESS_EQUAL) {
         SWAP(const byte*, data1, data2);
-        SWAP(uint, len1, len2);
+        SWAP(size_t, len1, len2);
         if (op == MP_BINARY_OP_LESS) {
             op = MP_BINARY_OP_MORE;
         } else {
             op = MP_BINARY_OP_MORE_EQUAL;
         }
     }
-    uint min_len = len1 < len2 ? len1 : len2;
+    size_t min_len = len1 < len2 ? len1 : len2;
     int res = memcmp(data1, data2, min_len);
     if (op == MP_BINARY_OP_EQUAL) {
         // If we are checking for equality, here're the answer
@@ -166,7 +191,7 @@ bool mp_seq_cmp_bytes(mp_uint_t op, const byte *data1, mp_uint_t len1, const byt
 
 // Special-case comparison function for sequences of mp_obj_t
 // Don't pass MP_BINARY_OP_NOT_EQUAL here
-bool mp_seq_cmp_objs(mp_uint_t op, const mp_obj_t *items1, mp_uint_t len1, const mp_obj_t *items2, mp_uint_t len2) {
+bool mp_seq_cmp_objs(mp_uint_t op, const mp_obj_t *items1, size_t len1, const mp_obj_t *items2, size_t len2) {
     if (op == MP_BINARY_OP_EQUAL && len1 != len2) {
         return false;
     }
@@ -174,7 +199,7 @@ bool mp_seq_cmp_objs(mp_uint_t op, const mp_obj_t *items1, mp_uint_t len1, const
     // Let's deal only with > & >=
     if (op == MP_BINARY_OP_LESS || op == MP_BINARY_OP_LESS_EQUAL) {
         SWAP(const mp_obj_t *, items1, items2);
-        SWAP(uint, len1, len2);
+        SWAP(size_t, len1, len2);
         if (op == MP_BINARY_OP_LESS) {
             op = MP_BINARY_OP_MORE;
         } else {
@@ -182,8 +207,8 @@ bool mp_seq_cmp_objs(mp_uint_t op, const mp_obj_t *items1, mp_uint_t len1, const
         }
     }
 
-    mp_uint_t len = len1 < len2 ? len1 : len2;
-    for (mp_uint_t i = 0; i < len; i++) {
+    size_t len = len1 < len2 ? len1 : len2;
+    for (size_t i = 0; i < len; i++) {
         // If current elements equal, can't decide anything - go on
         if (mp_obj_equal(items1[i], items2[i])) {
             continue;
@@ -215,11 +240,11 @@ bool mp_seq_cmp_objs(mp_uint_t op, const mp_obj_t *items1, mp_uint_t len1, const
 }
 
 // Special-case of index() which searches for mp_obj_t
-mp_obj_t mp_seq_index_obj(const mp_obj_t *items, mp_uint_t len, mp_uint_t n_args, const mp_obj_t *args) {
+mp_obj_t mp_seq_index_obj(const mp_obj_t *items, size_t len, size_t n_args, const mp_obj_t *args) {
     mp_obj_type_t *type = mp_obj_get_type(args[0]);
-    mp_obj_t *value = args[1];
-    uint start = 0;
-    uint stop = len;
+    mp_obj_t value = args[1];
+    size_t start = 0;
+    size_t stop = len;
 
     if (n_args >= 3) {
         start = mp_get_index(type, len, args[2], true);
@@ -228,19 +253,19 @@ mp_obj_t mp_seq_index_obj(const mp_obj_t *items, mp_uint_t len, mp_uint_t n_args
         }
     }
 
-    for (mp_uint_t i = start; i < stop; i++) {
+    for (size_t i = start; i < stop; i++) {
         if (mp_obj_equal(items[i], value)) {
             // Common sense says this cannot overflow small int
             return MP_OBJ_NEW_SMALL_INT(i);
         }
     }
 
-    nlr_raise(mp_obj_new_exception_msg(&mp_type_ValueError, "object not in sequence"));
+    mp_raise_ValueError("object not in sequence");
 }
 
-mp_obj_t mp_seq_count_obj(const mp_obj_t *items, mp_uint_t len, mp_obj_t value) {
-    mp_uint_t count = 0;
-    for (uint i = 0; i < len; i++) {
+mp_obj_t mp_seq_count_obj(const mp_obj_t *items, size_t len, mp_obj_t value) {
+    size_t count = 0;
+    for (size_t i = 0; i < len; i++) {
          if (mp_obj_equal(items[i], value)) {
               count++;
          }
