@@ -22,3 +22,9 @@ print(x)
 print(x[1:])
 print(x[:-1])
 print(x[2:3])
+
+# unsupported type on RHS of add
+try:
+    [] + None
+except TypeError:
+    print('TypeError')
