@@ -1,7 +1,11 @@
 try:
-    from collections import namedtuple
+    try:
+        from collections import namedtuple
+    except ImportError:
+        from ucollections import namedtuple
 except ImportError:
-    from _collections import namedtuple
+    print("SKIP")
+    raise SystemExit
 
 T = namedtuple("Tup", ["foo", "bar"])
 # CPython prints fully qualified name, what we don't bother to do so far
@@ -20,6 +24,9 @@ for t in T(1, 2), T(bar=1, foo=2):
 
     print(isinstance(t, tuple))
 
+# Create using positional and keyword args
+print(T(3, bar=4))
+
 try:
     t[0] = 200
 except TypeError:
@@ -66,7 +73,12 @@ T3 = namedtuple("TupComma", "foo bar")
 t = T3(1, 2)
 print(t.foo, t.bar)
 
-# Try single string with comma field seperator
+# Try tuple
+T4 = namedtuple("TupTuple", ("foo", "bar"))
+t = T4(1, 2)
+print(t.foo, t.bar)
+
+# Try single string with comma field separator
 # Not implemented so far
 #T2 = namedtuple("TupComma", "foo,bar")
 #t = T2(1, 2)
