@@ -5,14 +5,22 @@ import sys
 print(sys.__name__)
 print(type(sys.path))
 print(type(sys.argv))
-print(sys.version[:3])
-print(sys.version_info[0], sys.version_info[1])
 print(sys.byteorder in ('little', 'big'))
-print(sys.maxsize > 100)
-print(sys.implementation.name in ('cpython', 'micropython'))
 
 try:
-    sys.exit()
+    print(sys.maxsize > 100)
+except AttributeError:
+    # Effectively skip subtests
+    print(True)
+
+try:
+    print(sys.implementation.name in ('cpython', 'micropython'))
+except AttributeError:
+    # Effectively skip subtests
+    print(True)
+
+try:
+    raise SystemExit
 except SystemExit as e:
     print("SystemExit", e.args)
 
