@@ -18,8 +18,16 @@ print(spi)
 
 spi.init(SPI.SLAVE, phase=1)
 print(spi)
+try:
+    # need to flush input before we get an error (error is what we want to test)
+    for i in range(10):
+        spi.recv(1, timeout=100)
+except OSError:
+    print("OSError")
 
 spi.init(SPI.MASTER)
 spi.send(1, timeout=100)
 print(spi.recv(1, timeout=100))
 print(spi.send_recv(1, timeout=100))
+
+spi.deinit()
